from contextlib import contextmanager
from datetime import datetime
import getpass
import logging
from multiprocessing import Process
import os
import platform
import signal
import sys
import time
import random

from pymongo.errors import ConnectionFailure

from mtq.log import MongoStream, MongoHandler
from mtq.utils import handle_signals, now, setup_logging, nulltime


class Worker(object):
    '''
    Should create a worker from MTQConnection.new_worker.
    '''
    def __init__(self, factory, queues=(), tags=(), priority=0,
                 poll_interval=1, exception_handler=None,
                 log_worker_output=False, silence=False, extra_lognames=()):
        self.name = '%s.%s' % (platform.node(), os.getpid())
        self.extra_lognames = extra_lognames

        self.queues = queues
        self.tags = tags
        self.priority = priority

        self._log_worker_output = log_worker_output
        self.factory = factory
        self.poll_interval = poll_interval

        self.logger = logging.getLogger('mq.Worker')

        self._current = None
        self._handler = exception_handler
        self._pre_call = None
        self._post_call = None
        self.silence = silence

        self.collection = self.factory.worker_collection

    worker_id = '-'

    @contextmanager
    def register(self):
        '''
        Internal
        Contextmanager, register the birth and death of this worker

        eg::
            with worker.register():
                # Work
        '''
        self.collection = self.factory.worker_collection

        self.worker_id = self.collection.insert({'name': self.name,
                                                 'host': platform.node(),
                                                 'system': platform.system(),
                                                 'pid': os.getgid(),
                                                 'user': getpass.getuser(),
                                                 'started':now(),
                                                 'finished':datetime.fromtimestamp(0),
                                                 'check-in':datetime.fromtimestamp(0),
                                                 'working':True,
                                                 'queues': self.queues,
                                                 'tags': self.tags,
                                                 'log_output': bool(self._log_worker_output),
                                                 'terminate': False,
                                                 'terminate_status': 0,
                                                 })
        if self._log_worker_output:
            hdlr = MongoHandler(self.factory.logging_collection, {'worker_id':self.worker_id})
            self.logger.addHandler(hdlr)
        try:
            yield self.worker_id
        finally:
            if self._log_worker_output:
                self.logger.removeHandler(hdlr)

            query = {'_id': self.worker_id}
            update = {'$set':{'finished':now(), 'working':False}}
            self.collection.update(query, update)

    def check_in(self):

        query = {'_id': self.worker_id}
        update = {'$set':{'check-in':now(), 'working':True}}
        worker_info = self.collection.find_and_modify(query, update)
        if worker_info:
            should_exit = worker_info.get('terminate', False)
            status = worker_info.get('terminate_status', 0)
            return should_exit, status
        return False, 0

    def work(self, one=False, batch=False, failed=False, fail_fast=False):
        '''
        Main work function

        :param one: wait for the first job execute and then exit
        :param batch: work until the queue is empty, then exit
        '''
        with self.register():
            try:
                self.start_main_loop(one, batch, failed, fail_fast)
            except KeyboardInterrupt:
                self.logger.exception(None)
                if not self._current:
                    return

                self.logger.warn('Warm shutdown requested')
                proc, job = self._current
                proc.join(timeout=job.doc.get('timeout'))
                return

    def pop_item(self, pop_failed=False):
        job = self.factory.pop_item(worker_id=self.worker_id,
                                    queues=self.queues,
                                    tags=self.tags,
                                    failed=pop_failed,
                                    )
        return job

    def start_main_loop(self, one=False, batch=False, pop_failed=False, fail_fast=False, max_retries=10):
        '''
        Start the main loop and process jobs
        '''
        self.logger.info('Starting Main Loop mongo-host=%s mongo-db=%s' % (self.factory.db.connection.host,
                                                                           self.factory.db.name))
        self.logger.info('Starting Main Loop worker=%s _id=%s' % (self.name, self.worker_id))
        self.logger.info('Listening for jobs queues=[%s] tags=[%s]' % (', '.join(self.queues), ', '.join(self.tags)))
        retries = 0
        while 1:
            try:
                should_exit, status = self.check_in()
                if should_exit:
                    self.logger.info("Shutdown Requested (from DB)")
                    raise SystemExit(status)

                job = self.pop_item(pop_failed=pop_failed)
                if job is None:
                    if batch: break
                    time.sleep(self.poll_interval)
                    continue

                self.process_job(job)

                if one: break

                self.logger.info('Listening for jobs queues=[%s] tags=[%s]' % (', '.join(self.queues), ', '.join(self.tags)))
                retries = 0

            # Handle connection errors
            except ConnectionFailure as err:
                if fail_fast:
                    raise
                elif (retries == max_retries):
                    self.logger.exception('Retry limit reached (%d)', max_retries)
                    raise
                elif retries < max_retries:
                    retries += 1
                    self.logger.exception(err)
                    sleep_time = ((2 ** retries) + random.random()) * 0.1
                    self.logger.warn('Retrying in %.2f seconds', sleep_time)
                    time.sleep(sleep_time)
                else:
                    self.logger.exception(err)
                if one:
                    break
            except Exception as err:
                if fail_fast:
                    raise
                else:
                    self.logger.exception(err)
                if one:
                    break

        self.logger.info('Exiting Main Loop')

    def process_job(self, job):
        '''
        Process a single job in a multiprocessing.Process
        '''
        self.logger.info('Popped Job _id=%s queue=%s tags=%s' % (job.id, job.qname, ', '.join(job.tags)))
        self.logger.info(job.call_str)

        proc = Process(target=self._process_job, args=(job,))
        self._current = proc, job
        proc.start()
        timeout = job.doc.get('timeout')
        if timeout:
            self.logger.info("Job started, timing out after %s seconds" % timeout)
        else:
            self.logger.info("Job started, no time out")

        proc.join(timeout=job.doc.get('timeout'))
        if proc.is_alive():
            self.logger.error('Timeout occurred: interrupting job')
            os.kill(proc.pid, signal.SIGALRM)
            # Give the process 2 min to finish
            proc.join(timeout=min(job.doc.get('timeout'), 2 * 60))

            if proc.is_alive():
                self.logger.error('Process did not shut down after interrupt: terminating job')
                proc.terminate()

        self._current = None

        failed = proc.exitcode != 0

        if failed:
            self.logger.error('Job %s failed' % (job.doc['_id']))
        else:
            self.logger.info('Job %s finished successfully' % (job.doc['_id']))

        job.set_finished(failed)

        return  failed

    def _process_job(self, job):
        '''
        '''
        handle_signals()

        with setup_logging(self.factory.logging_collection, job.id):
            try:
                self._pre(job)
                job.apply()
            except:
                if self._handler:
                    exc_type, exc_value, traceback = sys.exc_info()
                    self._handler(job, exc_type, exc_value, traceback)
                raise
            finally:
                self._post(job)

    def _pre(self, job):
        if self._pre_call: self._pre_call(job)

    def _post(self, job):
        if self._post_call: self._post_call(job)

    def set_pre(self, func):
        self._pre_call = func

    def set_post(self, func):
        self._post_call = func

    def push_exception_handler(self, handler):
        self._handler = handler

    @property
    def num_backlog(self):
        'number of tasks this worker has to complete'
        return self.factory._items_cursor(queues=self.queues,
                                          tags=self.tags,
                                          ).count()


class WorkerProxy(object):
    'This is a representation of an actual worker process'

    def __init__(self, factory, doc):
        self.factory = factory
        self.doc = doc

    @property
    def id(self):
        return self.doc['_id']

    @property
    def name(self):
        return self.doc['name']

    @property
    def qnames(self):
        return self.doc.get('queues', ())

    @property
    def tags(self):
        return self.doc['tags']

    @property
    def num_processed(self):
        'number of tasks this worker has completed'
        collection = self.factory.queue_collection
        return collection.find({'worker_id': self.id}).count()

    @property
    def num_backlog(self):
        'number of tasks this worker has to complete'
        return self.factory._items_cursor(queues=self.qnames,
                                          tags=self.tags,
                                          ).count()

    @property
    def last_check_in(self):
        'last check in time'
        return self.doc.get('check-in', nulltime())

    def stream(self):
        collection = self.factory.logging_collection

        return MongoStream(collection,
                           doc={'worker_id': self.id},
                           finished=self.finished)

    def finished(self):
        'test if this worker is finished'
        coll = self.factory.worker_collection
        cursor = coll.find({'_id': self.id, 'working': False})
        return bool(cursor.count())



'''
Run a MTQ schedular or schedule a task

Schedule rule based on the iCal RFC 
    * http://labix.org/python-dateutil#head-e987b581aebacf25c7276d3e9214385a12a091f2
    * http://www.ietf.org/rfc/rfc2445.txt

Examples:

Run a single task now:

 mtq-scheduler --now mymodule.myfunction

Schedule a task to be run every hour:

 mtq-scheduler --add --task mymodule.myfunction --rule 'FREQ=HOURLY' 

Run a schedule server **should only ever be one running!**:
    
 mtq-scheduler --serve-forever   
'''

from __future__ import print_function
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from dateutil.rrule import rrulestr
from bson.objectid import ObjectId
from mtq.utils import config_dict
import mtq
import logging
from mtq.log import ColorStreamHandler


def test_rule(rule_str):
    try:
        _ = rrulestr(rule_str)
    except ValueError as err:
        raise SystemExit("Bad rule:" + err.message)


def pprint_scheduler(scheduler):
    rules = scheduler.rules
    if not rules.count():
        print('No Scheduled tasks')
        return 
    for rule in rules:
        print('%(_id)s | %(rule)30s | %(task)20s | %(queue)s' % rule)


def main():

    logger = logging.getLogger('mtq')
    logger.setLevel(logging.INFO)
    hdlr = ColorStreamHandler()
    logger.addHandler(hdlr)

    parser = ArgumentParser(description=__doc__, version='Mongo Task Queue (mtq) v%s' % mtq.__version__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-c', '--config', help='Python module containing MTQ settings.')
    
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--add', help='Schedule a new job', action='store_true')
    group.add_argument('-n', '--now', help='Run a job now!', metavar='TASK')
    group.add_argument('--remove', help='Remove a rule', type=ObjectId)
    group.add_argument('-u', '--update', help='Update a rule', type=ObjectId)
    group.add_argument('-l', '--list', help='List rules', action='store_true')
    group.add_argument('-s', '--serve-forever', '--run', help='List rules', action='store_true',
                       dest='run')
    parser.add_argument('-r', '--rule', help='Schedule rule based on the iCal RFC (http://www.ietf.org/rfc/rfc2445.txt)')
    parser.add_argument('-t', '--task', help='importable string of the task to be run. Must be a callable object with no arguments')
    parser.add_argument('-q', '--queue', help='name of the queue (default: default)')
    parser.add_argument('--tags', help='tag the job with these tags')
    parser.add_argument('--timeout', type=int, default=None,
                        help='Timeout after N seconds', metavar='N')
    args = parser.parse_args()
    
    config = config_dict(args.config)
    factory = mtq.from_config(config)
        
    scheduler = factory.scheduler()
    
    if args.rule:
        test_rule(args.rule)
        
    if args.add:
        if args.task is None:
            raise Exception('must specify task')
        _id = scheduler.add_job(args.rule, args.task, args.queue or 'default', args.tags, args.timeout)
        print('Added new scheduled task _id=%s' % (_id,))
    if args.update:
        scheduler.update_job(args.update, args.rule, args.task, args.queue, args.tags)
    elif args.list:
        pprint_scheduler(scheduler)
    elif args.remove:
        scheduler.remove_job(args.remove)
        print('Remove scheduled task _id=%s' % (args.remove,))
    elif args.now:
        queue = factory.queue(args.queue, tags=args.tags)
        queue.enqueue_call(args.now, timeout=args.timeout)
    elif args.run:
        scheduler.run()
    
        
        

if __name__ == '__main__':
    main()



'''
Control workers and jobs
'''
from __future__ import print_function
from argparse import ArgumentParser
from mtq.connection import MTQConnection
from mtq.utils import config_dict, now
from bson import ObjectId
from time import mktime
import mtq

def working(conn, args):
    coll = conn.worker_collection
    query = {'working':True}
    print('Bouncing %i workers' % coll.find(query).count())
    coll.update(query, {'$set':{'working':False}}, multi=True)
    print('Done')

def failed(conn, args):
    coll = conn.queue_collection
    query = {'failed':True}

    if args.func:
        query['execute.func_str'] = args.func

    if args.id:
        query['_id'] = args.id

    print('Flagging %i jobs as not failed' % coll.find(query).count())
    coll.update(query, {'$set':{'failed':False}}, multi=True)

def finish(conn, args):
    coll = conn.queue_collection
    query = {'finished':False}

    if args.id:
        query['_id'] = args.id
    print('Flagging %i jobs as finished' % coll.find(query).count())
    n = now()
    coll.update(query, {'$set':{'finished':True,
                                'finished_at': n,
                                'finished_at_': mktime(n.timetuple())}}, multi=True)

def shutdown(conn, args):
    coll = conn.worker_collection
    query = {'working':True}

    if args.all:
        pass
    if args.hostname:
        query['host'] = args.hostname
    if args.name:
        query['name'] = args.name

    if args.id:
        query['_id'] = args.id

    print('Shutting down %i workers' % coll.find(query).count())
    update = {'$set':{'terminate':True,
                      'terminate_status':args.status,
                      'working': False}}
    coll.update(query, update, multi=True)
    print('Done')


def main():

    parser = ArgumentParser(description=__doc__, version='Mongo Task Queue (mtq) v%s' % mtq.__version__)

    parser.add_argument('-c', '--config', help='Python module containing MTQ settings.')

    sp = parser.add_subparsers()

    wparser = sp.add_parser('working',
                            help=('Set all workers as not working. '
                                  'Live workers will reset working status'
                                  'on next check-in'))
    wparser.set_defaults(main=working)

    fparser = sp.add_parser('failed', help='Flag failed jobs as fixed')
    fparser.set_defaults(main=failed)
    group = fparser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action='store_true',
                       help='All jobs')
    group.add_argument('-f', '--func',
                       help='All jobs with function name')
    group.add_argument('-i', '--id', type=ObjectId,
                       help='Job Id')

    fparser = sp.add_parser('finish', help='Flag unfinished jobs as finished (use with caution)')
    fparser.set_defaults(main=finish)
    fparser.add_argument('-i', '--id',
                         type=ObjectId, required=True,
                         help='Job Id')

    sparser = sp.add_parser('shutdown',
                            help=('Schedule workers for shutdown. '
                                  'Workers will self-terminate on next check-in'))
    sparser.add_argument('-s', '--status', type=int,
                       help='Status code', default=1)

    sparser.set_defaults(main=shutdown)

    group = sparser.add_mutually_exclusive_group(required=True)

    group.add_argument('-a', '--all', action='store_true',
                       help='All workers')
    group.add_argument('-k', '--hostname',
                       help='Workers only running on host')
    group.add_argument('-n', '--name',
                       help='Worker name')
    group.add_argument('-i', '--id', type=ObjectId,
                       help='Worker Id')

    args = parser.parse_args()

    config = config_dict(args.config)
    factory = MTQConnection.from_config(config)

    args.main(factory, args)

if __name__ == '__main__':
    main()



'''
Created on Aug 1, 2013

@author: sean
'''
import signal
import traceback
import sys
import logging
from datetime import datetime
from bson.errors import InvalidId
from bson.objectid import ObjectId
from contextlib import contextmanager
import io
import pytz
from mtq import errors

class ImportStringError(Exception):
    pass

is_py3 = lambda: sys.version_info.major >= 3
def is_str(obj):
    if is_py3():
        return isinstance(obj, str)
    else:
        return isinstance(obj, basestring)

def is_unicode(obj):
    if is_py3():
        return isinstance(obj, str)
    else:
        return isinstance(obj, unicode)


def handle_signals():
    '''
    Handle signals in multiprocess.Process threads
    '''
    def handler(signum, frame):
        signal.signal(signal.SIGINT, signal.default_int_handler)

    def raise_timeout(signum, frame):
        raise errors.Timeout()

    def term_handler(signum, frame):
        traceback.print_stack(frame)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        raise SystemExit(-signum)

    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, term_handler)
    signal.signal(signal.SIGALRM, raise_timeout)

def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    For better debugging we recommend the new :func:`import_module`
    function to be used instead.

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    # force the import name to automatically convert to strings
    if is_unicode(import_name):
        import_name = str(import_name)
    try:
        if ':' in import_name:
            module, obj = import_name.split(':', 1)
        elif '.' in import_name:
            module, obj = import_name.rsplit('.', 1)
        else:
            return __import__(import_name)
        # __import__ is not able to handle unicode strings in the fromlist
        # if the module is a package
        if is_unicode(obj) and not is_py3():
            obj = obj.encode('utf-8')
        try:
            return getattr(__import__(module, None, None, [obj]), obj)
        except (ImportError, AttributeError):
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            modname = module + '.' + obj
            __import__(modname)
            return sys.modules[modname]
    except ImportError as e:
        if not silent:
            raise (ImportStringError(import_name, e), None, sys.exc_info()[2])

def ensure_capped_collection(db, collection_name, size_mb):
    '''
    '''
    if collection_name not in db.collection_names():
        db.create_collection(collection_name, capped=True,
                          size=(1024.**2) * size_mb)  # Mb

    return db[collection_name]


@contextmanager
def stream_logging(silence=False):
    from mtq.log import IOStreamLogger

    stdout = sys.stdout
    sys.stdout = IOStreamLogger(sys.stdout, silence)

    stderr = sys.stderr
    sys.stderr = IOStreamLogger(sys.stderr, silence)
    yield sys.stdout, sys.stderr

    sys.stdout = stdout
    sys.stderr = stderr

class UnicodeFormatter(logging.Formatter):
    def format(self, record):
        msg = logging.Formatter.format(self, record)
        # FIXME: should I be doing errors='replace'?
        if hasattr(msg, 'decode'):
            msg = msg.decode(errors='replace')
        return msg

mgs_template = """Job %s exited with exception:
Job Log:
   | %s

"""
@contextmanager
def setup_logging(collection, job_id):
    """
    Set up logging for worker.
    All log messages will be captured and saved into a document in the
    specified database collection.
    Additionally, the `job` logger will record start/end messages,
    with full logs exported only in the case of failure.
    """

    # create a handler that will capture specified logs in memory
    record = io.StringIO()
    record_hndlr = logging.StreamHandler(record)
    record_hndlr.setFormatter(UnicodeFormatter())
    record_hndlr.setLevel(logging.INFO)

    # configure root logger to capture all messages
    rootLogger = logging.getLogger()
    rootLogger.setLevel(logging.INFO)
    rootLogger.addHandler(record_hndlr)

    job_log = logging.getLogger('job')
    job_log.setLevel(logging.INFO)
    job_log.info('Starting Job %s' % job_id)

    try:
        # Run the job
        yield
    except Exception as exc:
        # Record the stack trace, then stop recording
        rootLogger.exception(exc)
        rootLogger.removeHandler(record_hndlr)

        # Dump saved logs into the job log
        text = record.getvalue().replace('\n', '\n   | ')
        msg = mgs_template % (job_id, text,)
        job_log.exception(msg)
        raise
    else:
        # Just record success in job log
        job_log.info("Job %s finished successfully" % (job_id,))
        rootLogger.removeHandler(record_hndlr)
    finally:
        log_entry = {
            'job_id': job_id,
            'message': record.getvalue(),
            'timestamp': now()
        }
        collection.insert(log_entry)



def now():
    now = datetime.utcnow()
    return now.replace(tzinfo=pytz.utc)

def nulltime():
    dt = datetime.utcfromtimestamp(0)
    return dt.replace(tzinfo=pytz.utc)

def config_dict(filename):
    config = {}
    if filename:
        return vars(import_string(filename))
    return config



def object_id(oid):
    try:
        return ObjectId(oid)
    except InvalidId:
        raise TypeError()


def wait_times(conn):
    coll = conn.queue_collection
    wait = { '$avg': { '$subtract':['$started_at_', '$enqueued_at_'] } }
    raw = coll.aggregate([{'$match':{'processed':True}}, {'$group':{'_id':'$qname', 'wait': wait } } ], cursor={})
    result = list(raw)
    return {item['_id']:item['wait'] for item in result}

def job_stats(conn, group_by='$execute.func_str', since=None):
    coll = conn.queue_collection
    duration = { '$avg': { '$subtract':['$finished_at_', '$started_at_'] } }
    wait = { '$avg': { '$subtract':['$started_at_', '$enqueued_at_'] } }
    count = {'$sum': 1}
    failed = {'$sum': {'$cmp':['$failed', False]}}
    queues = {'$addToSet': '$qname'}
    tags = {'$addToSet': '$push'}
    earliest = {'$min': '$finished_at'}
    latest = {'$max': '$finished_at'}

    match = {'$match':{'finished':True}}

    if since:
        match['$match']['finished_at'] = {'$gt':since}

    raw = coll.aggregate([match, {'$group':{'_id':group_by,
                                           'duration': duration,
                                           'wait_in_queue': wait,
                                           'count': count,
                                           'queues': queues,
                                           'tags': tags,
                                           'failed': failed,
                                           'latest': latest,
                                           'earliest': earliest,
                                            } }
                          ], cursor={})

    result = list(raw)

    return {item.pop('_id'):item for item in result}




def shutdown_worker(conn, worker_id=None):
    coll = conn.worker_collection
    query = {}
    if worker_id:
        query['_id'] = worker_id
    else:
        query = {'working':True}

    print(coll.update(query, {'$set':{'terminate':True}}, multi=True))

def last_job(conn, worker_id):
    coll = conn.queue_collection
    cursor = coll.find({'worker_id': worker_id}).sort('enqueued_at', -1)
    doc = next(cursor, None)

    if doc:
        from mtq.job import Job
        return Job(conn, doc)







"""
Starts an MTQ worker.
"""
from argparse import ArgumentParser

from mtq.connection import MTQConnection
from mtq.log import ColorStreamHandler
from mtq.utils import config_dict, object_id
import logging
import mtq

logger = logging.getLogger('worker')


def aux(args):
    logger = logging.getLogger('mq.Worker')
    logger.setLevel(logging.INFO)
    hdlr = ColorStreamHandler()
    logger.addHandler(hdlr)

    config = config_dict(args.config)

    tags = config.get('TAGS', ()) or args.tags
    queues = config.get('QUEUES', ()) or args.queues

    factory = MTQConnection.from_config(config)
    worker = factory.new_worker(queues=queues, tags=tags, log_worker_output=args.log_output,
                                poll_interval=args.poll_interval, args=args)

    if args.backlog:
        print(worker.num_backlog)
        return

    if config.get('exception_handler'):
        worker.push_exception_handler(config['exception_handler'])
    if config.get('pre_call'):
        worker._pre_call = config.get('pre_call')
    if config.get('post_call'):
        worker._post_call = config.get('post_call')

    if args.job_id:
        job = factory.get_job(args.job_id)
        if job is None:
            worker.logger.error('No job %s' % args.job_id)
            return

        worker.process_job(job)
        return

    worker.work(one=args.one, batch=args.batch, failed=args.failed)


def main():
    parser = ArgumentParser(description=__doc__, version='Mongo Task Queue (mtq) v%s'
                            % mtq.__version__, add_help=False)
    parser.add_argument('-c', '--config', help='Python module containing MTQ settings.')

    args, _ = parser.parse_known_args()
    config = config_dict(args.config)
    add_extra_arguments = config.get('extra_arguments')
    init_config = config.get('init')

    parser = ArgumentParser(description=__doc__,
                            version='Mongo Task Queue (mtq) v%s' % mtq.__version__, add_help=True)
    parser.add_argument('-c', '--config', help='Python module containing MTQ settings.')

    parser.add_argument('queues', nargs='*', default=['default'], help='The queues to listen on (default: %(default)r)')
    parser.add_argument('-r', '--reloader', action='store_true', help='Reload the worker when it detects a change')
    parser.add_argument('-p', '--poll-interval', help='Sleep interval to check for jobs', default=3, type=int)
    parser.add_argument('-t', '--tags', nargs='*', help='only process jobs which contain all of the tags', default=[])
    parser.add_argument('-l', '--log-output', action='store_true', help='Store job and woker ouput in the db, seealso mtq-tail')
    parser.add_argument('-1', '--one', action='store_true',
                        help='Process only the first job')
    parser.add_argument('--backlog', '--list-jobs', action='store_true',
                        help='List backlog of jobs and exit', dest='backlog')
    parser.add_argument('-b', '--batch', action='store_true',
                        help='Process jobs until the queue is empty, then exit')
    parser.add_argument('-j', '--job-id', type=object_id,
                        help='Process the job (even if it has already been processed)')
    parser.add_argument('-f', '--failed', action='store_true',
                        help='Process failed jobs')

    if add_extra_arguments:
        add_extra_arguments(parser)

    args = parser.parse_args()

    if init_config:
        init_config(args)

    if args.reloader:
        from werkzeug.serving import run_with_reloader
        run_with_reloader(lambda: aux(args))
    else:
        aux(args)


if __name__ == '__main__':
    main()



'''
Print stats on workers, queues and jobs
'''
from __future__ import print_function, division
from argparse import ArgumentParser
from mtq.connection import MTQConnection
from mtq.utils import config_dict, wait_times, last_job, now, job_stats
from datetime import timedelta, datetime

def reltime(dt):
    delta = now() - dt
    s = delta.total_seconds()

    hours = s / 3600.
    if hours > 1.5:
        return '%.1f hours ago' % hours

    minutes = s / 60.0
    if minutes > 1.5:
        return '%.1f minutes ago' % minutes
    return '%.1f seconds ago' % s

def queue_stats(factory, args):
    print('Queues:')
    latancy = wait_times(factory)
    for queue in factory.queues:
        tags = queue.all_tags
        print(' * Name: %s' % (queue.name))
        print('   Tags: [%s]' % (', '.join(tags)))
        print('   Count: %i' % queue.count)
        print('   Latancy: %.2f seconds' % latancy.get(queue.name, -1))

t = lambda tm:  (tm.ctime(), reltime(tm))

def worker_stats(factory, args):
    print('Workers:')
    for worker in factory.workers:
        print(' * %-10s' % (worker.name))
        print('   + Last Checked In: %s (%s)' % t(worker.last_check_in))
        print('   + Num-Processed: %i' % worker.num_processed)
        print('   + Backlog: %i' % worker.num_backlog)
        print('   + Queues:[%s]' % ', '.join(worker.qnames))
        print('   + Tags:[%s]' % ', '.join(worker.tags))
        job = last_job(factory, worker.id)
        if job:
            print('   + Last Job:')
            print('      - Id: %s' % job.doc['_id'])
            print('      - Enqueued: %s (%s)' % t(job.enqueued))
            print('      - Started : %s (%s)' % t(job.started))

            if job.doc.get('finished'):
                print('      - Finished: %s (%s)' % t(job.doc['finished_at']))
            else:
                print('      - Finished:  *In-Progress*')
        else:
            print('   *** Worker has not processed any jobs')

def print_job_stats(factory, args):

    stats = job_stats(factory, since=args.since)
    for key, value in stats.items():
        print('+', key)
        for item in value.items():
            if isinstance(item[1], datetime):
                item = item[0], '%s (%s)' % t(item[1])
            else:
                item = item[0], repr(item[1])

            print('   - %s: %s' % item)

    cursor = factory.queue_collection.find({'finished':False})
    if cursor.count():
        print()
        print('Unfinished Jobs:')
        for job in cursor:
            if job['processed']:
                print(' * %s %s -- started %s, enqueued %s' % (job['execute']['func_str'], job['_id'], reltime(job['started_at']), reltime(job['enqueued_at'])))
            else:
                print(' * %s %s -- **waiting**, enqueued %s' % (job['execute']['func_str'], job['_id'], reltime(job['enqueued_at'])))

    cursor = factory.queue_collection.find({'failed':True})
    if cursor.count():
        print()
        print('Failed Jobs:')
        for job in cursor:
            print(' *', job['execute']['func_str'], job['_id'])


def max_age(arg):
    if arg.lower().endswith('s'):
        return now() - timedelta(0, int(arg[:-1]))
    elif arg.lower().endswith('m'):
        return now() - timedelta(0, int(arg[:-1]) * 60)
    elif arg.lower().endswith('h'):
        return now() - timedelta(0, int(arg[:-1]) * 60 * 60)
    elif arg.lower().endswith('d'):
        return now() - timedelta(int(arg[:-1]))
    else:
        return now() - timedelta(0, int(arg))

def print_db_stats(factory, args):
    import pymongo
    print('PyMongo', pymongo.version)
    print('Mongo database', factory.db.name)
    for collection_names in factory.db.collection_names():
        stats = factory.db.command('collStats', collection_names, scale=1024)
        if stats.get('capped'):
            print(" * %(ns)s (%(count)i items)" % stats)
            print("   Using %(size)s of %(storageSize)s Kb" % stats,
                  '%.2f%%' % (100 * stats['size'] / stats['storageSize']),
                  '(capped)' if stats.get('capped') else '',
                  )


def main():

    parser = ArgumentParser(description=__doc__, version='0.0')
    parser.add_argument('-c', '--config', help='Python module containing MTQ settings.')
    parser.add_argument('-m', '--max-age',
                        help='Maximum age of jobs (e.g. 1d) unit may be one of s (seconds), m (minutes), h (hours) or d (days) ',
                        type=max_age, dest='since', metavar='AGE',
                        default=None)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-q', '--queues', action='store_const',
                       const=queue_stats, dest='action',
                       help='print stats on queues')
    group.add_argument('-w', '--worker', action='store_const', const=worker_stats, dest='action',
                       help='print stats on workers')
    group.add_argument('-j', '--jobs', action='store_const', const=print_job_stats, dest='action',
                       help='print stats on jobs')
    group.add_argument('-s', '--storage', '--db', action='store_const', const=print_db_stats, dest='action',
                       help='print stats on database')

    args = parser.parse_args()

    config = config_dict(args.config)
    factory = MTQConnection.from_config(config)
    args.action(factory, args)

if __name__ == '__main__':
    main()



'''
MQ module
'''
from .connection import MTQConnection
from . import defaults
from .queue import Queue
from .worker import Worker, WorkerProxy
from .job import Job
from .schedule import Scheduler
from mtq.defaults import _task_map

try:
    from _version import __version__
except ImportError:
    __version__ = '0.7.1'


def default_connection():
    '''
    Create a default mtq instance to created queues, workers, and jobs
    '''
    return MTQConnection.default()


def create_connection(db,
                      collection_base=defaults._collection_base,
                      qsize=defaults._qsize,
                      workersize=defaults._workersize,
                      logsize=defaults._logsize,
                      extra_lognames=()):
    '''
    Create a new mtq instance to created queues, workers, and jobs

    :param db: mongo database
    :param collection_base: base name for collection
    :param qsize: the size of the capped collection of the queue

    '''
    return MTQConnection(db, collection_base, qsize, workersize, logsize, extra_lognames)


def from_config(config=None, client=None):
    '''
    Create a new mtq instance to created queues, workers, and jobs

    :param config: configutation dict, with the parameters
        * DB_HOST
        * DB
        * COLLECTION_BASE
        * COLLECTION_SIZE
    :param client: a pymongo.MongoClient or None
    '''
    return MTQConnection.from_config(config, client)


def task(func=None, name=None):

    def inner(func_inner):
        task_name = name or '%s.%s' % (func_inner.__module__, func_inner.__name__)
        _task_map[task_name] = func_inner

    if func is None:
        return inner
    else:
        return inner(func)



'''
Created on Aug 2, 2013

@author: sean
'''
from mtq.utils import import_string, now, nulltime
from mtq.log import MongoStream
from bson.objectid import ObjectId
from time import mktime

class Job(object):
    '''
    A Job is just a convenient datastructure to pass around job (meta) data.
    
    Do not create directly, use MTQConnection.get_job
    '''
    def __init__(self, factory, doc):
        self.factory = factory
        self.doc = doc
    def __repr__(self):
        return '<job queue=%r tags=%r func_name=%r>' % (self.qname, self.tags, self.func_name)

    @property
    def tags(self):
        'List of tags for this job'
        return self.doc['tags']

    @property
    def qname(self):
        'The name of the queue that this job is in'
        return self.doc['qname']

    @property
    def func_name(self):
        'The name of the task to execute'
        return self.doc['execute']['func_str']

    @property
    def id(self):
        'the identifier for this job'
        return self.doc['_id']


    @property
    def func(self):
        'a callable function for workers to execute'

        if self.func_name in self.factory._task_map:
            return self.factory._task_map[self.func_name]

        return import_string(self.func_name)

    @property
    def call_str(self):
        args = [repr(arg) for arg in self.args]
        args.extend('%s=%r' % item for item in self.kwargs.items())
        args = ', '.join(args)
        return '%s(%s)' % (self.func_name, args)

    @property
    def enqueued(self):
        return self.doc['enqueued_at']

    @property
    def started(self):
        return self.doc['started_at']

    @property
    def args(self):
        'The arguments to call func with'
        return self.doc['execute']['args']

    @property
    def kwargs(self):
        'The keyword arguments to call func with'
        return self.doc['execute']['kwargs']

    def apply(self):
        'Execute this task syncronusly'
        return self.func(*self.args, **self.kwargs)

    def set_finished(self, failed=False):
        '''
        Mark this jog as finished.
        
        :param failed: if true, this was a failed job
        '''
        n = now()

        update = {'$set':{'processed':True,
                          'failed':failed,
                          'finished':True,
                          'finished_at': n,
                          'finished_at_': mktime(n.timetuple())
                          }
                  }

        self.factory.queue_collection.update({'_id':self.id}, update)

        if not failed:
            data = self.factory.queue_collection.find_one({'_id':self.id})
            if data:
                self.factory.queue_collection.remove({'_id':self.id})
                self.factory.finished_jobs_collection.insert(data)

    def stream(self):
        '''
        Get a stream to read log lines from this job  
        '''
        return MongoStream(self.factory.logging_collection,
                           doc={'job_id': self.id},
                           finished=self.finished)

    def finished(self):
        '''
        test if this job has finished
        '''
        collection = self.factory.queue_collection
        cursor = collection.find({'_id':self.id, 'processed':True})
        return bool(cursor.count())

    def cancel(self):
        self.set_finished()


    @classmethod
    def new(cls, name, tags, priority, execute, timeout, mutex=None):

        n = now()
        no = mktime(n.timetuple())
        return {
               'qname':name,
               'tags': tags,
               'process_after': n,
               'priority': priority,

               'execute': execute,
               'enqueued_at': n,
               'enqueued_at_': no,
               'started_at': nulltime(),
               'started_at_': 0.0,
               'finished_at': nulltime(),
               'finished_at_': 0.0,
               'processed': False,
               'failed': False,
               'finished': False,
               'timeout':timeout,
               'worker_id': ObjectId('000000000000000000000000'),
               'mutex': mutex,
               }







'''
Created on Aug 1, 2013

@author: sean
'''
from mtq.utils import now
import io
import logging
import sys
import time

class ColorStreamHandler(logging.Handler):
    '''
    Handler to write to stdout with colored output
    '''
    WARNING = '\033[93m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = "\033[1m"
    COLOR_MAP = {'ERROR': '%s%s[%%s]%s' % (BOLD, FAIL, ENDC),
                 'WARNING': '%s%s[%%s]%s' % (BOLD, WARNING, ENDC),
                 'DEBUG': '%s%s[%%s]%s' % (BOLD, OKBLUE, ENDC),
                 'INFO': '%s%s[%%s]%s' % (BOLD, OKBLUE, ENDC),
                 }

    def color_map(self, header, level):
        return self.COLOR_MAP.get(level, '[%s]') % header

    def __init__(self, level=logging.INFO):
        logging.Handler.__init__(self, level=level)
    def emit(self, record):
        header = ''
        if record.levelno == logging.INFO:
            header = record.name
            message = record.getMessage()
            stream = sys.stdout
        else:
            stream = sys.stderr
            if record.exc_info:
                err = record.exc_info[1]
                header = '----%s:%s' % (type(err).__name__, record.name)
                if err.args:
                    message = err.args[0]
                else:
                    message = str(err)
            else:
                header = record.name
                message = record.getMessage()

        if stream.isatty() and not sys.platform.startswith('win'):
            header = self.color_map(header, record.levelname)
        stream.write('%s %s\n' % (header, message))


class BSONFormatter(object):
    '''
    format recort to bison dict
    '''
    def __init__(self, *args, **extra_tags):
        object.__init__(self, *args)
        self.extra_tags = extra_tags

    def format(self, record):
        if isinstance(record.msg, dict):
            data = record.msg
        elif isinstance(record.msg, (list, tuple)):
            data = {'items': record.msg}
        else:
            data = {'message':'%s\n' % (record.msg,)}

        data.update(logLevel=record.levelname, logModule=record.module, logName=record.name, **self.extra_tags)
        return data

class MongoHandler(logging.Handler):
    '''
    Log to monog db
    '''
    def __init__(self, collection, doc):

        logging.Handler.__init__(self, logging.INFO)

        self.collection = collection
        self.doc = doc
        self.setFormatter(BSONFormatter())


    def emit(self, record):
        doc = self.doc.copy()
        data = self.format(record)
        doc.update(data)
        self.collection.insert(doc)

class TextIOWrapperSmart(io.TextIOWrapper):
    '''
    allow print to io.TextIOWrapper
    '''
    def write(self, s):
        if isinstance(s, bytes):
            s = s.decode()
        return io.TextIOWrapper.write(self, s)

class IOStreamLogger(logging.Logger):
    def __init__(self, stream, silence=False):

        self.stream_name = name = getattr(stream, 'name', None)
        logging.Logger.__init__(self, name, level=logging.INFO)
        self.stream = stream
        self.silence = silence

        hdlr = logging.StreamHandler(stream)
        hdlr.setLevel(logging.INFO)
        self.addHandler(hdlr)

    def readable(self):
        return 'r' in self.stream.mode

    @property
    def closed(self):
        return self.stream.closed

    def writable(self):
        return 'w' in self.stream.mode or 'a' in self.stream.mode

    def seekable(self):
        return False

    def write(self, message):
        self.info(message)
        return len(message)

    def flush(self):
        self.stream.flush()


class MongoStream(object):
    '''
    File like object to read/write to mongodb
    '''
    def __init__(self, collection, doc, stream=None, finished=None, silence=False):
        self.collection = collection
        self.doc = doc
        self.stream = stream
        self._finished = finished
        self.silence = silence
        self.stream_name = getattr(stream, 'name', None)

    def readable(self):
        return False

    @property
    def closed(self):
        return False

    def writable(self):
        return True

    def seekable(self):
        return False

    def write(self, message):
        doc = self.doc.copy()
        doc.update(message=message, time=now())

        self.collection.insert(doc)

        if self.stream and not self.silence:
            self.stream.write(message)
        return len(message)

    def flush(self):
        pass

    def loglines(self, follow=False):
        starting_finished = not follow or self._finished()

        cursor = self.collection.find(self.doc,
                                      await_data=not starting_finished,
                                      tailable=not starting_finished)
        while 1:
            for item in cursor:
                text = item.get('message', '')
                yield text

            if starting_finished or self._finished():
                return

            time.sleep(1)

        return

def mstream(collection, doc, stream=None, silence=False):
    '''
    Create a buffered mongo stream (good for print statements)
    '''
    return TextIOWrapperSmart(MongoStream(collection, doc, sys.stdout, silence), line_buffering=True)



def add_all(loggers, handlers):
    for l in loggers:
        for h in handlers:
            l.addHandler(h)

def remove_all(loggers, handlers):
    for l in loggers:
        for h in handlers:
            l.addHandler(h)






# -*- coding: utf-8 -*-
"""
Compatibility layer for writing pymongo3.x looking code that works on both
pymongo2.8 and 3.x
"""
# Third party imports
import bson
import pymongo


# Constants
# ----------------------------------------------------------------------------
AS_CLASS = 'pymongo3compat_as_class'
FIND_LIKE = 'pymongo3compat_find_like_method'
PYMONGO_28 = pymongo.version_tuple == (2, 8)
PYMONGO_3 = pymongo.version_tuple[0] == 3
RENAME_MAP = {'filter': 'spec',
              'projection': 'fields',
              'allow_partial_results': 'partial'}


# Error types
# ----------------------------------------------------------------------------
class Error(Exception):
    pass


class PyMongoVersionError(Error):
    pass


class PyMongoMigrationHelperError(Error):
    pass


# Helpers
# ----------------------------------------------------------------------------
class _CodecOptions:
    """
    Mock CodecOption object for use with pymongo 2.8.

    http://api.mongodb.org/python/3.0/api/bson/codec_options.html#bson.codec_options.CodecOptions
    """
    def __init__(self, document_class=None):
        self.document_class = document_class


def _mongo_client(*args, **kwargs):
    """
    Helper function for use with pymongo 2.8 that returns either a MongoClient
    or a MongoReplicaSetClient.

    In pymongo 3.x there is a single client.

    http://api.mongodb.org/python/3.0/changelog.html#mongoclient-changes
    """
    if getattr(kwargs, 'replicaSet', None):
        MongoClient = pymongo.mongo_replica_set_client.MongoReplicaSetClient
    else:
        MongoClient = pymongo.mongo_client.MongoClient

    return MongoClient(*args, **kwargs)


if PYMONGO_28:
    CodecOptions = _CodecOptions
    MongoClient = _mongo_client
elif PYMONGO_3:
    CodecOptions = bson.codec_options.CodecOptions
    MongoClient = pymongo.mongo_client.MongoClient
else:
    raise PyMongoVersionError


def _fix_kwargs(kwargs):
    """
    Adapt the kwargs so that renamed keywords are passed to the method
    according to the correct version.

    http://api.mongodb.org/python/3.0/changelog.html#collection-changes
    """
    for new_key in RENAME_MAP:
        if new_key in kwargs:
            old_key = RENAME_MAP[new_key]
            kwargs[old_key] = kwargs.pop(new_key)
    return kwargs


def _find_like_method(*args, **kwargs):
    """
    Helper method that adapts the find_like methods so that code looks like
    pymongo3.x but can run on both 2.8 and 3.x series
    """
    method = kwargs.pop(FIND_LIKE, None)
    collection = kwargs.pop('collection', None)
    func = getattr(collection, method, None)

    if not (collection and method and func):
        raise PyMongoMigrationHelperError

    if PYMONGO_28:
        kwargs = _fix_kwargs(kwargs)

        # Any non existing attributes will be returned as a Collection
        as_class = getattr(collection, AS_CLASS)
        if not isinstance(as_class, pymongo.collection.Collection):
            kwargs['as_class'] = as_class
            setattr(collection, AS_CLASS, None)  # Reset value for collection
        cursor_or_document = func(*args, **kwargs)
    elif PYMONGO_3:
        cursor_or_document = func(*args, **kwargs)
    else:
        raise PyMongoVersionError

    return cursor_or_document


# Wrapper functions
# ----------------------------------------------------------------------------
def find(*args, **kwargs):
    """
    Wraps the find method making the syntax similar to pymongo3.x.
    """
    kwargs[FIND_LIKE] = 'find'
    return _find_like_method(*args, **kwargs)


def find_and_modify(*args, **kwargs):
    """
    Wraps the find_and_modify method making the syntax similar to pymongo3.x.
    """
    kwargs[FIND_LIKE] = 'find_and_modify'
    return _find_like_method(*args, **kwargs)


def find_one(*args, **kwargs):
    """
    Wraps the find_one method making the syntax similar to pymongo3.x.
    """
    kwargs[FIND_LIKE] = 'find_one'
    return _find_like_method(*args, **kwargs)


def with_options(codec_options=None, collection=None):
    """
    Helper method for using a document_class (as_class) in pymongo 2.8.

    http://api.mongodb.org/python/3.0/api/pymongo/collection.html#pymongo.collection.Collection.with_options
    """
    if not (collection and codec_options):
        raise PyMongoMigrationHelperError

    if PYMONGO_28:
        if codec_options is not None:
            as_class = getattr(codec_options, 'document_class', None)
            setattr(collection, AS_CLASS, as_class)
        else:
            raise PyMongoMigrationHelperError
    elif PYMONGO_3:
        collection = collection.with_options(codec_options=codec_options)
    else:
        raise PyMongoVersionError

    return collection



import unittest
import pymongo
import mtq
import logging

log = logging.getLogger('mtq.test')
log.setLevel(logging.INFO)


def test_func(*args, **kwargs):
    log.info('Running test_func')
    return args, kwargs


def test_func_fail(*args, **kwargs):
    log.warning('Raising a test exception in test_func_fail')
    raise Exception()


class MTQTestCase(unittest.TestCase):
    connection = None

    def setUp(self):
        if type(self).connection is None:
            type(self).connection = pymongo.MongoClient()

        self.db = self.connection['mtq_testing']
        self.factory = mtq.create_connection(db=self.db)

    def tearDown(self):
        # uncomment to dump database state after each test
        # from pprint import pprint
        # for coll in self.db.collection_names():
        #     print('\n%s:' % coll)
        #     pprint(list(self.db[coll].find()))
        self.connection.drop_database('mtq_testing')



'''
Created on Aug 5, 2013

@author: sean
'''
import unittest
from os.path import dirname

def main():
    import coverage
    cov = coverage.coverage()
    cov.start()
    
    loader = unittest.loader.TestLoader()
    tests = loader.discover(dirname(__file__))
    runner = unittest.TextTestRunner()
    runner.run(tests) 
    cov.stop()
    cov.report(omit=['**/site-packages/**', '**/tests/**'])
    cov.html_report(omit=['**/site-packages/**', '**/tests/**'])

if __name__ == '__main__':
    main()



'''
Created on Aug 4, 2013

@author: sean
'''
from mtq.tests.fixture import MTQTestCase
import mtq.tests.fixture
from mtq.utils import now
import unittest
from mtq.queue import QueueError
import time

class TestQueue(MTQTestCase):

    def test_create_queue(self):
        'Creating queues.'
        q = self.factory.queue('my-queue', ['tags'])
        self.assertEqual(q.name, 'my-queue')
        self.assertEqual(q.count, 0)

        self.assertEqual(str(q), 'my-queue')
        self.assertEqual(repr(q), "<mtq.Queue name:my-queue tags:('tags',)>")

    def test_enqueue_call(self):
        'Creating queues.'
        q = self.factory.queue('my-queue', ['tags'])
        q.enqueue_call('call-me')
        self.assertEqual(q.count, 1)

        job = q.enqueue_call(mtq.tests.fixture.test_func, args=(1,))
        self.assertEqual(job.func_name, 'mtq.tests.fixture.test_func')
        self.assertEqual(q.count, 2)

        job = q.enqueue_call(mtq.tests.fixture.test_func, args=(1,), kwargs={'this':1})
        self.assertEqual(job.func_name, 'mtq.tests.fixture.test_func')
        self.assertEqual(q.count, 3)

    def test_enqueue_call_error(self):
        'Creating queues.'
        q = self.factory.queue('my-queue', ['tags'])

        with self.assertRaises(QueueError):
            q.enqueue_call(object())

        with self.assertRaises(TypeError):
            q.enqueue_call('fine', args='str')

        with self.assertRaises(TypeError):
            q.enqueue_call('fine', kwargs='str')


    def test_enqueue(self):
        'Creating queues.'
        q = self.factory.queue('my-queue', ['tags'])

        q.enqueue('call-me')


    def test_jobs(self):
        'Creating queues.'
        q = self.factory.queue('my-queue', ['tags'])
        q.enqueue_call('call-me')
        self.assertEqual(len(q.jobs), 1)
        job = q.jobs[0]
        self.assertEqual(job.tags, ['tags'])
        self.assertEqual(job.func_name, 'call-me')
        self.assertEqual(job.qname, 'my-queue')

        job.cancel()
        self.assertEqual(len(q.jobs), 0)

    def test_is_empty(self):
        """Detecting empty queues."""
        q = self.factory.queue('example')
        self.assertEqual(q.is_empty(), True)

        self.factory.queue_collection.insert({'qname':'example',
                                              'process_after': now(),
                                              'priority': 0,
                                              'processed':False
                                              })
        self.assertEqual(q.is_empty(), False)

    def test_all_tags(self):

        q = self.factory.queue('my-queue')
        q.enqueue_call('call-me', tags=['1', '2'])
        q.enqueue_call('call-me', tags=['2', '3'])

        self.assertEqual(q.count, 2)

        self.assertEqual(q.all_tags, set(['1', '2', '3']))



    def test_pop(self):
        q = self.factory.queue('my-queue')
        q.enqueue_call('call-me1', tags=['1', '2'])

        job = q.pop('worker')
        self.assertIsNotNone(job)
        self.assertEqual(job.func_name, 'call-me1')

        job = q.pop('worker')
        self.assertIsNone(job)

    def test_mutex1(self):
        q = self.factory.queue('my-queue')
        q.enqueue_call('call-me1', mutex={'key': 'key1', 'count': 1})
        q.enqueue_call('call-me2', mutex={'key': 'key1', 'count': 1})
        q.enqueue_call('call-me3', mutex={'key': 'key2', 'count': 1})

        job1 = q.pop('worker')
        job2 = q.pop('worker')
        job3 = q.pop('worker')

        self.assertIsNotNone(job1)
        self.assertIsNotNone(job2)
        self.assertIsNone(job3)

        job1.set_finished()
        job2.set_finished()

        job3 = q.pop('worker')

        self.assertIsNotNone(job3)

    def test_mutex2(self):

        q = self.factory.queue('my-queue')
        q.enqueue_call('call-me1', mutex={'key': 'key1', 'count': 2})
        q.enqueue_call('call-me2', mutex={'key': 'key1', 'count': 2})
        q.enqueue_call('call-me3', mutex={'key': 'key1', 'count': 2})
        q.enqueue_call('call-me4', mutex={'key': 'key2', 'count': 1})

        job1 = q.pop('worker')
        job2 = q.pop('worker')
        job3 = q.pop('worker')
        job4 = q.pop('worker')

        self.assertIsNotNone(job1)
        self.assertIsNotNone(job2)
        self.assertIsNotNone(job3)
        self.assertIsNone(job4)



if __name__ == '__main__':
    unittest.main()



'''
Created on Aug 5, 2013

@author: sean
'''

from mtq.tests.fixture import MTQTestCase
import mtq
import unittest
import mock

from pymongo.errors import ConnectionFailure


TEST_JOB_DATA = {'execute':{'func_str':'mtq.tests.fixture.test_func_fail',
                            'args':(1, 2), 'kwargs':{}},
                            '_id':'abc',
                            'qname':'qname',
                            'tags':[],
                        }

class TestWorker(MTQTestCase):

    def test_create_worker(self):
        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'])

        self.assertEqual(worker.queues, ['q1', 'q2'])
        self.assertEqual(worker.tags, ['t1', 't2'])

    def test_register_worker(self):
        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'])

        with worker.register():
            workers = list(self.factory.workers)
            self.assertEqual(len(workers), 1)
            workerpxy = workers[0]

            workerpxy.name

            self.assertEqual(workerpxy.qnames, ['q1', 'q2'])
            self.assertEqual(workerpxy.tags, ['t1', 't2'])
            self.assertEqual(workerpxy.num_processed, 0)
            self.assertFalse(workerpxy.finished())

        self.assertTrue(workerpxy.finished())

        workers = list(self.factory.workers)
        self.assertEqual(len(workers), 0)

    def read_log(self, job_id):
        entries = list(self.db.mq.log.find({'job_id': job_id}))
        self.assertEqual(len(entries), 1)
        entry = entries[0]
        self.assertIn('message', entry)
        return entry['message']

    @mock.patch('mtq.worker.handle_signals')
    @mock.patch('mtq.tests.fixture.test_func')
    def test_sync_process(self, test_func, handle_signals):
        job = mtq.Job(self.factory, {'execute':{'func_str':'mtq.tests.fixture.test_func',
                                                'args':(1, 2), 'kwargs':{}},
                                     '_id':'abc',
                                     'qname':'qname',
                                     })

        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'])

        worker._process_job(job)
        handle_signals.assert_called_once_with()
        test_func.assert_called_once_with(1, 2)

        msg = self.read_log(job_id='abc')
        self.assertIn('Starting Job', msg)
        self.assertIn('finished successfully', msg)

        pre = mock.Mock('pre')
        post = mock.Mock('post')
        worker.set_pre(pre)
        worker.set_post(post)

        worker._process_job(job)
        pre.assert_called_once_with(job)
        post.assert_called_once_with(job)

    @mock.patch('mtq.worker.handle_signals')
    @mock.patch('mtq.tests.fixture.test_func')
    def test_sync_process_failure(self, test_func, handle_signals):
        job = mtq.Job(self.factory, {'execute':{'func_str':'mtq.tests.fixture.test_func_fail',
                                                'args':(1, 2), 'kwargs':{}},
                                     '_id':'abc2',
                                     'qname':'qname',
                                     })

        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'])
        pre = mock.Mock('pre')
        post = mock.Mock('post')
        worker.set_pre(pre)
        worker.set_post(post)

        with self.assertRaises(Exception):
            worker._process_job(job)

        msg = self.read_log(job_id='abc2')
        self.assertIn('Starting Job', msg)
        self.assertIn('Exception', msg)
        self.assertNotIn('finished successfully', msg)

        pre.assert_called_once_with(job)
        post.assert_called_once_with(job)

        exc_handler = mock.Mock('exc_handler')
        worker.push_exception_handler(exc_handler)

        with self.assertRaises(Exception):
            worker._process_job(job)

        self.assertEqual(exc_handler.call_count, 1)

    def test_process(self):
        job = mtq.Job(self.factory, {'execute':{'func_str':'mtq.tests.fixture.test_func',
                                                'args':(1, 2), 'kwargs':{}},
                                     '_id':'abc3',
                                     'qname':'qname',
                                     'tags':[],
                                     })

        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'])

        failed = worker.process_job(job)
        self.assertFalse(failed)

    def test_process_error(self):
        job = mtq.Job(self.factory, {'execute':{'func_str':'mtq.tests.fixture.test_func_fail',
                                                'args':(1, 2), 'kwargs':{}},
                                     '_id':'abc4',
                                     'qname':'qname',
                                     'tags':[],
                                     })

        worker = self.factory.new_worker(['q1', 'q2'], ['t1', 't2'], silence=True)

        failed = worker.process_job(job)
        self.assertTrue(failed)


    def test_tags(self):

        worker = self.factory.new_worker(['q1', 'q2'], ['linux-64', 'hostname:host1'], silence=True)
        q = self.factory.queue('q1')
        self.assertIsNone(worker.pop_item())

        q.enqueue_call('test')
        self.assertIsNotNone(worker.pop_item())

        q.enqueue_call('test', tags=['linux-64'])
        item = worker.pop_item()
        self.assertIsNotNone(item)
        self.assertEqual(item.tags, ['linux-64'])

        q.enqueue_call('test', tags=['hostname:host1'])
        item = worker.pop_item()
        self.assertIsNotNone(item)
        self.assertEqual(item.tags, ['hostname:host1'])

        q.enqueue_call('test', tags=['linux-64', 'hostname:host1'])
        item = worker.pop_item()
        self.assertIsNotNone(item)
        self.assertEqual(set(item.tags), set(['linux-64', 'hostname:host1']))

        q.enqueue_call('test', tags=['linux-64', 'hostname:host2'])
        self.assertIsNone(worker.pop_item())

        q.enqueue_call('test', tags=['linux-65'])
        self.assertIsNone(worker.pop_item())

        q.enqueue_call('test', tags=[])
        self.assertIsNotNone(worker.pop_item())

        self.assertIsNone(worker.pop_item())

    def test_unexpected_error_fail_fast(self):

        worker = self.factory.new_worker(['q1'], ['linux-64'], silence=True)
        worker.pop_item = mock.Mock()
        worker.pop_item.side_effect = KeyError("This is expected")

        with self.assertRaises(KeyError):
            worker.start_main_loop(fail_fast=True)

    def test_exponential_backoff_on_connection_error(self):
        worker = self.factory.new_worker(['q1'], ['linux-64'], silence=True)
        worker.pop_item = mock.Mock()
        worker.pop_item.side_effect = ConnectionFailure("This is expected")
        with mock.patch('logging.Logger.exception') as mock_logger:
            with self.assertRaises(ConnectionFailure):
                worker.start_main_loop(max_retries=2)
            self.assertEqual(mock_logger.call_count, 3)
            mock_logger.assert_called_with('Retry limit reached (%d)', 2)

    def test_unexpected_error(self):

        worker = self.factory.new_worker(['q1'], ['linux-64'], silence=True)
        worker.pop_item = mock.Mock()
        worker.pop_item.return_value = None

        def side_effect(*_, **__):
            worker.pop_item.side_effect = None
            worker.pop_item.return_value = None
            raise KeyError("This is expected")

        worker.pop_item.side_effect = side_effect

        worker.start_main_loop(batch=True, fail_fast=False)


if __name__ == '__main__':
    unittest.main()



'''
Created on Aug 5, 2013

@author: sean
'''

from mtq.tests.fixture import MTQTestCase
import mtq.tests.fixture
from mtq.utils import now
import unittest
from mtq.queue import QueueError

class TestQueue(MTQTestCase):

    def test_create_job(self):
        'Creating jobs.'
        q = self.factory.queue('my-queue', ['tags'])
        q.enqueue(mtq.tests.fixture.test_func, 1, 2, a=1, b=2)
        job = q.pop('worker')

        self.assertEqual(job.func_name, 'mtq.tests.fixture.test_func')

        job = self.factory.get_job(job.id)

        self.assertEqual(job.args, [1, 2])
        self.assertEqual(job.kwargs, {'a':1, 'b':2})
        self.assertEqual(job.func, mtq.tests.fixture.test_func)

        args, kwargs = job.apply()
        self.assertEqual(args, (1, 2))
        self.assertEqual(kwargs, {'a':1, 'b':2})

        self.assertEqual(job.tags, ['tags'])

    def test_job_finished(self):
        'Creating jobs.'
        q = self.factory.queue('my-queue', ['tags'])
        job = q.enqueue(mtq.tests.fixture.test_func, 1, 2, a=1, b=2)
        self.assertFalse(job.finished())
        job = q.pop('worker')
        self.assertTrue(job.finished())

    def test_call_str(self):
        job = mtq.Job(self.factory, {'execute':{'func_str':'fs', 'args':(), 'kwargs':{}}})
        self.assertEqual(job.call_str, 'fs()')

        job = mtq.Job(self.factory, {'execute':{'func_str':'fs', 'args':(1, 2), 'kwargs':{}}})
        self.assertEqual(job.call_str, 'fs(1, 2)')

        job = mtq.Job(self.factory, {'execute':{'func_str':'fs', 'args':(1, 2), 'kwargs':{'a':3}}})
        self.assertEqual(job.call_str, 'fs(1, 2, a=3)')

        job = mtq.Job(self.factory, {'execute':{'func_str':'fs', 'args':(), 'kwargs':{'a':3}}})
        self.assertEqual(job.call_str, 'fs(a=3)')



if __name__ == '__main__':
    unittest.main()



'''
Created on Aug 5, 2013

@author: sean
'''
from dateutil.rrule import rrulestr
import time
from datetime import timedelta
from mtq.utils import now
import logging

from mtq.pymongo3compat import find_and_modify


class Scheduler(object):
    
    def __init__(self, factory):
        self.factory = factory
        
        self.logger = logging.getLogger('mtq.Scheduler')

    
    def add_job(self, rule, task, queue, tags=(), timeout=None):
        collection = self.factory.schedule_collection
        return collection.insert({'rule':rule, 'task':task, 'queue':queue, 'tags':tags,
                                  'paused':False, 'active':True,
                                  'modified':now(),
                                  'created':now(),
                                  'checked':now(),
                                  'timeout': timeout,
                                  })
 
    def remove_job(self, _id):
        collection = self.factory.schedule_collection
        return collection.remove({'_id':_id})
     

    def update_job(self, _id, rule=None, task=None, queue=None, tags=None):
        collection = self.factory.schedule_collection
        query = {'_id':_id}
        doc = {'$set':{'modified':now()}}
        if rule is not None:
            doc['$set']['rule'] = rule
        if task is not None:
            doc['$set']['task'] = task
        if queue is not None:
            doc['$set']['queue'] = queue
        if tags is not None:
            doc['$set']['tags'] = tags
             
        return collection.update(query, doc)
 
     
    @property
    def rules(self):
        collection = self.factory.schedule_collection
        return collection.find({'paused':False, 'active':True})

    def check_rule(self, doc, n):
        collection = self.factory.schedule_collection
        query = {'_id':doc['_id'], 'checked': doc['checked']}
        update = {'$set': {'checked': n}}
        return bool(find_and_modify(query, update, projection=[], collection=collection))
    

    def enqueue_from_rule(self, rule):
        queue = self.factory.queue(rule['queue'], tags=rule['tags'])
        queue.enqueue_call(rule['task'], timeout=rule.get('timeout'))
    
    def run(self):
        self.logger.info('Running Scheduler')
        poll_interval = 5
        try:
            while 1:
                
                next_event = now() + timedelta(1)
                 
                for rule in self.rules:
                    n = now() 
                    rrule = rrulestr(rule['rule'], dtstart=rule['created'])
                    items = rrule.between(rule['checked'], n)
                    if len(items) > 1:
                        self.logger.warn("Schedular missed %i tasks! Enqueuing latest" % (len(items)))
                    if items:
                        if self.check_rule(rule, n):
                            self.logger.info("Enqueueing task %r (%s)" % (rule['task'], items[0].ctime()))
                            self.enqueue_from_rule(rule)
                        else:
                            self.logger.warn("Another schedular has already run this task. moving on")
                    
                    next_event = min(next_event, rrule.after(n))
                
                self.logger.debug("Next event %s" % next_event.ctime())
                next_event = (next_event - now()).total_seconds()
                self.logger.debug("Next event in %i seconds" % next_event)
                
                sleep = max(1, min(next_event, poll_interval)) 
                self.logger.debug("Sleping for %i seconds" % sleep)
                time.sleep(sleep)
        
        except KeyboardInterrupt:
            self.logger.exception('Exiting main loop')



from __future__ import print_function

import time
from mtq.utils import stream_logging
import logging
import io

logger = logging.getLogger('job.foo.bar')
logger.setLevel(logging.INFO)

def raise_error():
    x = 1 + 1
    logger = logging.getLogger('job.foo.bar')
    logger.info("this is an info")
    logging.getLogger('woc').info("this is from the woc log")
    logging.getLogger('job').info("this is an info 2333")
    asdf

def do_something():
    logger.info("this is a log")
    print("this is a print")
    
def long_running_task():
    try:
        print("this is long_running_task ...")
        for _ in range(60):
            print("this is still a long_running_task ...")
            time.sleep(10)
    
    except Exception as err:
        print("This is an exception")
        raise err
    except BaseException as err:
        print("This is a base exception")
        raise err
    
def main():
    record = io.BytesIO()
    hndlr = logging.StreamHandler(record)
    logger = logging.getLogger('job')
    logger.setLevel(logging.INFO)
    logger.addHandler(hndlr)
    hndlr.setLevel(logging.INFO)
    logger.warn("asdfasdf")
    
    print(record)
    record.seek(0)
    print(record.read())


if __name__ == '__main__':
    main()


'''
Created on Aug 2, 2013

@author: sean
'''
import mtq
from pymongo.mongo_client import MongoClient
from mtq.defaults import _collection_base, _qsize, _workersize, _logsize, \
    _task_map
from mtq.utils import ensure_capped_collection, now
from time import mktime
from pymongo import ASCENDING
from mtq.pymongo3compat import find


class MTQConnection(object):
    '''
    Base object that you should use to create all other TQ objects

    ## Init

    :param db: mongo database
    :param collection_base: base name for collection
    :param qsize: the size of the capped collection of the queue

    ..sealso: MTQConnection.default, MTQConnection.from_config
    '''

    def __init__(self, db, collection_base=_collection_base, qsize=_qsize,
                 workersize=_workersize, logsize=_logsize, extra_lognames=()):
        self.db = db
        self.collection_base = collection_base
        self.qsize = qsize
        self.workersize = workersize
        self.logsize = logsize
        self._task_map = _task_map.copy()
        self.extra_lognames = extra_lognames

    def _destroy(self):
        'Destroy ALL data'
        self.db.connection.drop_database(self.db)

    @classmethod
    def default(cls):
        '''
        Create an MTQConnection default configuration using mongo from localhost
        '''
        return cls.from_config()

    @classmethod
    def from_config(cls, config=None, client=None):
        '''
        Create an MTQConnection from a config dict,

        :param config: configutation dict, with the parameters
            * DB_HOST
            * DB
            * COLLECTION_BASE
            * COLLECTION_SIZE
        :param client: a pymongo.MongoClient or None
        '''
        if config is None:
            config = {}

        if 'connection' in config:
            return config['connection']

        if client is None:
            client = MongoClient(config.get('DB_HOST', 'mongodb://localhost/?journal=true'), tz_aware=True)

        db = getattr(client, config.get('DB', 'mq'))

        base = config.get('COLLECTION_BASE', _collection_base)
        qsize = config.get('COLLECTION_SIZE', _qsize)
        return cls(db, base, qsize, extra_lognames=config.get('extra_lognames', ()))

    def job_stream(self, job_id):
        '''
        Get a file like object for the output of a job
        '''
        job = self.get_job(job_id)
        return job.stream()

    def worker_stream(self, worker_name=None,
                            worker_id=None):
        '''
        Get a file like object for the output of a worker
        '''
        worker = self.get_worker(worker_name, worker_id)
        return worker.stream()

    @property
    def queue_collection(self):
        'The collection to push jobs to'
        collection_name = '%s.queue' % (self.collection_base)
        return self.db[collection_name]

    @property
    def finished_jobs_collection(self):
        'The collection to push jobs to'
        collection_name = '%s.finished_jobs' % (self.collection_base)
        return ensure_capped_collection(self.db, collection_name, self.qsize)

    @property
    def logging_collection(self):
        'The collection to push log lines to'
        db = self.db
        collection_name = '%s.log' % self.collection_base
        return ensure_capped_collection(db, collection_name, self.logsize)

    @property
    def schedule_collection(self):
        'The collection to push log lines to'
        db = self.db
        collection_name = '%s.schedule' % self.collection_base
        return db[collection_name]

    def make_query(self, queues, tags, priority=0, processed=False, failed=False, **query):
        '''
        return a mongodb query dict to get the next task in the queue
        '''
        query.update({
                'priority':{'$gte':priority},
                'process_after': {'$lte':now()},
                 })
        if failed:
            query['failed'] = True
        elif processed is not None:
            query['processed'] = processed

        if queues:
            if len(queues) == 1:
                query['qname'] = queues[0]
            else:
                query['qname'] = {'$in': queues}

        query.update(self.make_tag_query(tags))
        return query

    def make_tag_query(self, tags):
        'Query for tags'
        if not tags:
            tag_query = {}
#         elif len(tags) == 1:
#             return {'tags': tags[0]}
        else:
            # tags on job must be a subset of jobs on the worker
            # i.e. assert job['tags'] in worker['tags']
            tag_query = {'tags': {'$not':{ '$elemMatch' : {'$nin': tags}}}}

        return tag_query


    def add_mutex(self, query):
        running_query = self.make_query(None, None, processed=True)
        cursor = find(running_query, projection={'mutex':1, '_id':0},
                      collection=self.queue_collection)

        if not cursor.count():
            return

        # Populate dictionary of mutex_key:count
        mutex = {}

        # Iterate over running jobs
        for item in cursor:
            item_mutex = item.get('mutex', {})
            if not item_mutex: continue
            mutext_key = item_mutex.get('key')
            if not mutext_key: continue
            mutex.setdefault(mutext_key, 0)
            mutex[mutext_key] += 1

        # Query should inclue jobs with no mutex key or where its key is not in any running jobs
        _or = [{'mutex': None}, {'mutex': {'$exists': False}}, {'mutex.key': {'$nin': mutex.keys()}}]

        # Query should inclue jobs where the mutex.count is > the # already running
        for key, already_running in mutex.items():
            _or.append({'mutex.key': key, 'mutex.count': {'$gt': already_running}})

        query['$or'] = _or

    def pop_item(self, worker_id, queues, tags, priority=0, failed=False):
        'Pop an item from the queue'
        n = now()
        update = {'$set':{'processed':True,
                          'started_at': n,
                          'started_at_': mktime(n.timetuple()),
                          'worker_id':worker_id}
                  }

        query = self.make_query(queues, tags, priority, failed)
        self.add_mutex(query)

        doc = self.queue_collection.find_and_modify(query, update, sort=[('enqueued_at', ASCENDING)])

        if doc is None:
            return None
        else:
            return mtq.Job(self, doc)

    def push_item(self, job_id):
        query = {'_id': job_id}
        update = {'$set':{'processed':False}}
        doc = self.queue_collection.find_and_modify(query, update)

    def _items_cursor(self, queues, tags, priority=0, processed=False, limit=None, reverse=False):
        query = self.make_query(queues, tags, priority, processed=processed)
        cursor = self.queue_collection.find(query)

        if reverse:
            cursor = cursor.sort('enqueued_at', -1)
        if limit:
            cursor = cursor.limit(limit)

        return cursor

    def items(self, queues, tags, priority=0, processed=False, limit=None, reverse=False):
        cursor = self._items_cursor(queues, tags, priority, processed, limit, reverse)
        return [mtq.Job(self, doc) for doc in cursor]

    def queue(self, name='default', tags=(), priority=0):
        '''
        Create a queue object

        :param name: the name of the queue
        :param tags: default tags to give to jobs
        :param priority: (not implemented yet)
        '''

        return mtq.Queue(self, name, tags, priority)



    def new_worker(self, queues=(), tags=(), priority=0, silence=False,
                   log_worker_output=False, poll_interval=3, args=None):
        '''
        Create a worker object

        :param queues: names of queues to pop from (these are OR'd)
        :param tags: jobs *must* have all these tags to be processed by this worker
        :param priority: (not implemented yet)
        :param log_worker_output: if true, log worker output to the db
        '''
        worker = mtq.Worker(self, queues, tags, priority,
                            log_worker_output=log_worker_output,
                            silence=silence, extra_lognames=self.extra_lognames, poll_interval=poll_interval)

        self.args = args
        self.worker = worker
        return worker

    #===========================================================================
    # Workers
    #===========================================================================
    @property
    def worker_collection(self):
        'Collection to register workers to'
        collection_name = '%s.workers' % self.collection_base
        return self.db[collection_name]

    @property
    def queues(self):
        'List of existing queues'
        collection = self.queue_collection
        qnames = collection.find().distinct('qname')
        return [mtq.Queue(self, qname) for qname in qnames]

    @property
    def workers(self):
        '''
        List of existing workers

        :returns: a WorkerProxy object
        '''
        collection = self.worker_collection
        return [mtq.WorkerProxy(self, item) for item in collection.find({'working':True})]


    def get_job(self, job_id):
        '''
        retrieve a job
        '''
        collection = self.queue_collection
        doc = collection.find_one({'_id':job_id})
        if doc is None:
            return None

        return mtq.Job(self, doc)

    def get_worker(self, worker_name=None, worker_id=None):
        '''
        retrieve a worker
        '''
        coll = self.worker_collection
        if worker_name:
            query = {'name':worker_name}
        elif worker_id:
            query = {'_id':worker_id}
        else:
            raise TypeError('must give one of worker_name or worker_id')
        doc = coll.find_one(query)
        if doc is None:
            raise TypeError('Could not find worker')

        return mtq.WorkerProxy(self, doc)


    #===========================================================================
    # Scheduler
    #===========================================================================

    def scheduler(self):
        return mtq.Scheduler(self)





'''
Created on Aug 1, 2013

@author: sean
'''
from argparse import ArgumentParser
from mtq.connection import MTQConnection
from bson.objectid import ObjectId
import sys



def main():
    
    parser = ArgumentParser(description=__doc__, version='0.0')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-j','--job-id', type=ObjectId)
    group.add_argument('-w','--worker-name')
    group.add_argument('--worker-id', type=ObjectId)
    
    parser.add_argument('-f', '--follow', action='store_true', 
                       help=('Dont stop when end of file is reached, but '
                             'rather to wait for additional data to be'
                             'appended to the input.'))
    
    args = parser.parse_args()
    
    factory = MTQConnection.from_config()
    
    if args.job_id:
        stream = factory.job_stream(args.job_id)
    elif args.worker_name or  args.worker_id:
        stream = factory.worker_stream(args.worker_name, args.worker_id)
        
    try:
        for line in stream.loglines(args.follow):
            sys.stdout.write(line)
    except KeyboardInterrupt:
        pass
    
        
    
if __name__ == '__main__':
    main()





from mtq.job import Job
from mtq.utils import is_str

class QueueError(Exception):
    pass

class Queue(object):
    '''
    A queue to enqueue an pop tasks
    
    Do not create directly use MTQConnection.queue 
    '''
    def __str__(self):
        return self.name
    def __repr__(self):
        return '<mtq.Queue name:%s tags:%r>' % (self.name, self.tags)

    def __init__(self, factory, name='default', tags=(), priority=0):

        self.name = name or 'default'
        self.factory = factory
        self.tags = tuple(tags) if tags else ()
        self.priority = priority


    def enqueue(self, func_or_str, *args, **kwargs):
        '''
        Creates a job to represent the delayed function call and enqueues
        it.
        
        Expects the function to call, along with the arguments and keyword
        arguments.
        
        The function argument `func_or_str` may be a function or a string representing the location of a function
        '''
        return self.enqueue_call(func_or_str, args, kwargs)

    def enqueue_call(self, func_or_str, args=(), kwargs=None, tags=(), priority=None, timeout=None, mutex=None):
        '''
        Creates a job to represent the delayed function call and enqueues
        it.
        
        It is much like `.enqueue()`, except that it takes the function's args
        and kwargs as explicit arguments.  Any kwargs passed to this function
        contain options for MQ itself.
        '''
        if not is_str(func_or_str):
            name = getattr(func_or_str, '__name__', None)
            module = getattr(func_or_str, '__module__', None)

            if not (name and module):
                raise QueueError('can not enqueue %r (type %r)' % (func_or_str, type(func_or_str)))

            func_or_str = '%s.%s' % (module, name)

        if args is None:
            args = ()
        elif not isinstance(args, (list, tuple)):
            raise TypeError('argument args must be a tuple')
        if kwargs is None:
            kwargs = {}
        elif not isinstance(kwargs, dict):
            raise TypeError('argument kwargs must be a dict')

        execute = {'func_str': func_or_str, 'args':tuple(args), 'kwargs': dict(kwargs)}

        if priority is None:
            priority = self.priority

        tags = self.tags + tuple(tags)
        doc = Job.new(self.name, tags, priority, execute, timeout, mutex)
        collection = self.factory.queue_collection
        collection.insert(doc)

        return Job(self.factory, doc)


    @property
    def count(self):
        'The number of jobs in this queue (filtering by tags too)'
        collection = self.factory.queue_collection
        query = self.factory.make_query([self.name], self.tags, self.priority)
        return collection.find(query).count()

    @property
    def num_failed(self):
        'The number of jobs in this queue (filtering by tags too)'
        collection = self.factory.queue_collection
        query = self.factory.make_query([self.name], self.tags, failed=True, processed=None)
        return collection.find(query).count()

    def is_empty(self):
        'The number of jobs in this queue (filtering by tags too)'
        return bool(self.count == 0)

    @property
    def all_tags(self):
        'All the unique tags of jobs in this queue'
        collection = self.factory.queue_collection
        return set(collection.find({'qname':self.name}).distinct('tags'))

    def pop(self, worker_id=None):
        'Pop a job off the queue'
        return self.factory.pop_item(worker_id, [self.name], self.tags, self.priority)

    @property
    def jobs(self):
        return self.factory.items([self.name], self.tags, self.priority)

    @property
    def finished_jobs(self):
        return self.factory.items([self.name], self.tags, self.priority,
                                  processed=True, limit=30, reverse=True)

    @property
    def all_jobs(self):
        return self.factory.items([self.name], self.tags, self.priority, processed=None, limit=30, reverse=True)

